from collections.abc import Callable
from types import FunctionType, MethodType
from typing import NamedTuple

InjectableType = type | FunctionType | MethodType


class ServiceInfo(NamedTuple):
    provides: type | None
    name: str | None


class ServiceDependency(NamedTuple):
    service: type
    name: str | None
    optional: bool = False


class ServiceDefinition(NamedTuple):
    provides: type
    factory: InjectableType
    dependencies: list[tuple[str, ServiceDependency]]
    initializers: list[Callable]
    finalizers: list[Callable]



import importlib
import importlib.util
import inspect
import os
import warnings
from pathlib import Path
from types import ModuleType, SimpleNamespace
from typing import Any

__all__ = ("Settings",)

SELVA_SETTINGS_MODULE = "SELVA_SETTINGS_MODULE"
DEFAULT_SELVA_SETTINGS_MODULE = "configuration/settings.py"

SELVA_ENV = "SELVA_ENV"


def is_valid_conf(conf: str) -> bool:
    """Checks if the config item can be collected into settings

    Config settings that are exported must start with an uppercase letter
    followed by other uppercase letters, numbers or underscores
    """

    if not (conf[0].isalpha() and conf[0].isupper()):
        return False

    return all((i.isalpha() and i.isupper()) or i.isnumeric() or i == "_" for i in conf)


def extract_valid_keys(settings: ModuleType) -> dict[str, Any]:
    """Collect settings from module into dict"""
    return {
        name: value
        for name, value in inspect.getmembers(settings)
        if is_valid_conf(name)
    }


def get_settings_for_env(env: str = None) -> dict[str, Any]:
    settings_module_path = Path(
        os.getenv(SELVA_SETTINGS_MODULE, DEFAULT_SELVA_SETTINGS_MODULE)
    )
    settings_module_path = settings_module_path.with_suffix(".py")

    if env is not None:
        settings_module_path = settings_module_path.with_stem(
            f"{settings_module_path.stem}_{env}"
        )

    settings_module_path = settings_module_path.absolute()

    try:
        spec = importlib.util.spec_from_file_location(
            "selva_settings", settings_module_path
        )
        settings_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(settings_module)
    except FileNotFoundError:
        # TODO: write tests
        warnings.warn(f"settings module not found: {settings_module_path}")
        return {}
    except ImportError:
        # TODO: find out what to do
        # TODO: write tests
        raise

    return extract_valid_keys(settings_module)


def get_settings() -> dict[str, Any]:
    settings = get_settings_for_env()

    if active_env := os.getenv(SELVA_ENV):
        env_settings = get_settings_for_env(active_env)
        settings |= env_settings

    return settings


class Settings(SimpleNamespace):
    def __init__(self):
        super().__init__(**get_settings())



import inspect
from typing import Annotated


class Named:
    def __init__(self):
        raise TypeError("Cannot instantiate Named")

    def __class_getitem__(cls, item: tuple[type, str]):
        if not isinstance(item, tuple) or len(item) != 2:
            raise ValueError("Named[...] requires 2 arguments")

        service, name = item

        if not inspect.isclass(service):
            raise ValueError(
                "Invalid argument for Named[...]. First argument must be a class"
            )

        if not isinstance(name, str):
            raise ValueError(
                "Invalid argument for Named[...]. Second argument must be str"
            )

        return Annotated[service, name]



from selva.web.errors import NotFoundError

from .path_converter import PathConverter


class StrPathConverter(PathConverter[str]):
    def from_path(self, value: str) -> str:
        return value

    def into_path(self, obj: str) -> str:
        return obj


class IntPathConverter(PathConverter[int]):
    def from_path(self, value: str) -> int | None:
        try:
            return int(value)
        except ValueError:
            raise NotFoundError()

    def into_path(self, obj: int) -> str:
        return str(obj)


class FloatPathConverter(PathConverter[float]):
    def from_path(self, value: str) -> float | None:
        try:
            return float(value)
        except ValueError:
            raise NotFoundError()

    def into_path(self, obj: float) -> str:
        return str(obj)


class BoolPathConverter(PathConverter[bool]):
    TRUE_VALUES = ["1", "true", "True", "TRUE"]
    FALSE_VALUES = ["0", "false", "False", "FALSE"]
    POSSIBLE_VALUE = TRUE_VALUES + FALSE_VALUES

    def from_path(self, value: str) -> bool | None:
        if value in self.POSSIBLE_VALUE:
            return value in self.TRUE_VALUES

        raise NotFoundError()

    def into_path(self, obj: bool) -> str:
        return str(obj)



from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable

from asgikit.responses import HttpResponse

from selva.di import service
from selva.web.request import RequestContext


class Middleware(ABC):
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        service(cls)

    @abstractmethod
    async def execute(
        self,
        chain: Callable[[RequestContext], Awaitable[HttpResponse | None]],
        context: RequestContext,
    ) -> HttpResponse | None:
        raise NotImplementedError()



from collections.abc import Awaitable
from typing import Generic, TypeVar

from selva.di import service
from selva.web import HttpResponse

__all__ = ("IntoResponse",)

T = TypeVar("T")


class IntoResponse(Generic[T]):
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        service(provides=cls.__orig_bases__[0])(cls)

    def into_response(self, value: T) -> HttpResponse | Awaitable[HttpResponse]:
        raise NotImplementedError()



from collections.abc import Awaitable
from typing import Generic, TypeVar

from selva.di import service

from .context import RequestContext

__all__ = ("FromRequest",)

T = TypeVar("T")


class FromRequest(Generic[T]):
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        service(provides=cls.__orig_bases__[0])(cls)

    def from_request(self, context: RequestContext) -> T | Awaitable[T]:
        raise NotImplementedError()



from collections.abc import Awaitable
from types import ModuleType
from typing import Any, Type, TypeVar

from selva.utils.maybe_async import maybe_async
from selva.utils.package_scan import scan_packages

from .decorators import DI_SERVICE_ATTRIBUTE
from .errors import (
    DependencyLoopError,
    ServiceNotFoundError,
    ServiceWithoutDecoratorError,
)
from .interceptor import Interceptor
from .service.model import InjectableType, ServiceDefinition, ServiceDependency
from .service.parse import get_dependencies, parse_definition
from .service.registry import ServiceRegistry

TService = TypeVar("TService")


class Container:
    def __init__(self):
        self.registry = ServiceRegistry()
        self.store: dict[tuple[type, str | None], Any] = {}
        self.finalizers: list[Awaitable] = []
        self.interceptors: list[Type[Interceptor]] = []

    def register(
        self,
        service: InjectableType,
        *,
        provides: type = None,
        name: str = None,
    ):
        definition = parse_definition(service, provides)
        provided_service = definition.provides
        self.registry[provided_service, name] = definition

    def service(self, service_type: type):
        service_info = getattr(service_type, DI_SERVICE_ATTRIBUTE, None)
        if not service_info:
            raise ServiceWithoutDecoratorError(service_type)

        provides, name = service_info
        self.register(service_type, provides=provides, name=name)

    def define_singleton(self, service_type: type, instance: Any, *, name: str = None):
        self.store[service_type, name] = instance

    def interceptor(self, interceptor: Type[Interceptor]):
        self.register(
            interceptor,
            provides=Interceptor,
            name=f"{interceptor.__module__}.{interceptor.__qualname__}",
        )
        self.interceptors.append(interceptor)

    def scan(self, *packages: str | ModuleType):
        def predicate(item: Any):
            return hasattr(item, DI_SERVICE_ATTRIBUTE)

        for service in scan_packages(packages, predicate):
            provides, name = getattr(service, DI_SERVICE_ATTRIBUTE)
            self.register(service, provides=provides, name=name)

    def has(self, service: type) -> bool:
        definition = self.registry.get(service)
        return definition is not None

    async def get(
        self,
        service_type: TService,
        *,
        context: Any = None,
        name: str = None,
        optional=False,
    ) -> TService:
        dependency = ServiceDependency(service_type, name=name, optional=optional)
        return await self._get(dependency, context)

    async def create(self, service: type, *, context: Any = None) -> Any:
        dependencies = await self._resolve_dependencies(service, context)
        instance = service(**dependencies)
        return instance

    async def _resolve_dependencies(self, service: InjectableType, context: Any = None):
        return {
            name: await self._get(dep, context)
            for name, dep in get_dependencies(service)
        }

    def _get_from_cache(self, service_type: type, name: str | None) -> Any | None:
        # try getting from singleton store
        if instance := self.store.get((service_type, name)):
            return instance

        return None

    async def _get(
        self,
        dependency: ServiceDependency,
        context: Any = None,
        stack: list = None,
    ) -> Any | None:
        service_type, name = dependency.service, dependency.name

        if instance := self._get_from_cache(service_type, name):
            return instance

        try:
            definition = self.registry[service_type, name]
        except ServiceNotFoundError:
            if dependency.optional:
                return None
            raise

        stack = stack or []

        if service_type in stack:
            raise DependencyLoopError(stack + [service_type])

        stack.append(service_type)
        instance = await self._create_service(definition, stack, name, context)
        stack.pop()

        return instance

    async def _create_service(
        self,
        definition: ServiceDefinition,
        stack: list[type],
        name: str = None,
        context: Any = None,
    ) -> Any:
        factory = definition.factory

        dependencies = {
            name: await self._get(dep, context, stack)
            for name, dep in definition.dependencies
        }

        # check if service have been created from initializers
        if instance := self._get_from_cache(definition.provides, name):
            return instance

        instance = await maybe_async(factory, **dependencies)

        self.store[definition.provides, name] = instance

        await self._run_initializer(definition, instance)
        self._setup_finalizers(definition, instance)

        if definition.provides is not Interceptor:
            await self._run_interceptors(instance, definition.provides)

        return instance

    async def _run_initializer(self, definition: ServiceDefinition, instance: Any):
        for initializer in definition.initializers:
            dependencies = await self._resolve_dependencies(initializer)
            await maybe_async(initializer, instance, **dependencies)

    def _setup_finalizers(self, definition: ServiceDefinition, instance: Any):
        for finalizer in definition.finalizers:
            self.finalizers.append(maybe_async(finalizer, instance))

    async def _run_interceptors(self, instance: Any, service_type: type):
        for cls in self.interceptors:
            interceptor = await self.get(
                Interceptor, name=f"{cls.__module__}.{cls.__qualname__}"
            )
            await maybe_async(interceptor.intercept, instance, service_type)

    async def run_finalizers(self):
        for finalizer in reversed(self.finalizers):
            await finalizer

        self.finalizers.clear()
